package okapi;

import okapi.net.protocol.ProcessResponse;
import okapi.net.protocol.OkapiNetMessage;
import okapi.net.protocol.NodeCleanup;
import okapi.net.protocol.ConnectMessage;
import okapi.net.protocol.ProcessRequest;
import okapi.net.protocol.OkapiMessageType;
import java.net.Socket;
import java.util.concurrent.atomic.AtomicInteger;

/**
 * A NetProcessConsumer belongs to a SchedulerServer, consuming processes by
 * sending them over the network to some form of Node, usually the standard
 * SlaveNode.
 * @author Sean Kelly
 */
public class NetProcessConsumer extends ProcessConsumer {
    private ConnectMessage connMsg; // Message sent by connecting node
    private AtomicInteger currLoad; // Number of jobs the node is handling
    private boolean alive; // Determines whether or not to grab more tasks
    
    private Socket conn; // Connection to the client Node
    private NetHandler commHandler; // Handles incoming communications from
                                    // assigned SlaveNode
    
    public NetProcessConsumer(SchedulerServer sched, Socket connection,
            ConnectMessage request)
    {
        super(sched);
        connMsg = request;
        conn = connection;
        
        alive = true;
        currLoad = new AtomicInteger(0);
    }
    
    @Override
    public boolean acceptJobs() {
        boolean ret = ((currLoad.get() - 1) <= connMsg.numJobs) && alive;
        return ret;
    }

    @Override
    public void consumeProcess(OkapiProcess proc) {
        currLoad.getAndIncrement();
        ProcessRequest req = new ProcessRequest();
        req.toDo = proc;
        // Send the job, the NetHandler should handle returning.
        OkapiNetMessage.sendMessage(req, conn);
    }

    @Override
    public void init() {
        // Start incoming network communcation handling
        commHandler = new NetHandler(conn, sched, currLoad);
        commHandler.start();
        
    }

    @Override
    public void cleanup() {
        kill();
        commHandler.kill();
        // TODO Send kill signal to SlaveNode
        NodeCleanup clnmsg = new NodeCleanup();
        OkapiNetMessage.sendMessage(clnmsg, conn);
    }
    
    @Override
    public void kill()
    {
        super.kill();
        alive = false;
    }

    /**
     * A NetHandler simply watches the given Socket for completed processes to
     * return to the ProcessConsumer.
     */
    private class NetHandler extends Thread
    {
        private Socket comm; // The socket with which to talk to the node
        private Scheduler sched; // scheduler to report finished jobs to
        
        private AtomicInteger currLoad; // Number of jobs the owning
                                        // Process consumer has
        private boolean alive; // Whether or not this thread should run
        
        public NetHandler(Socket conn, Scheduler schedu, AtomicInteger jb)
        {
            comm = conn;
            sched = schedu;
            
            currLoad = jb;
            alive = true;
        }
        
        @Override
        public void run()
        {
            OkapiNetMessage test;
            ProcessResponse resp;
            while(alive)
            {
                test = OkapiNetMessage.receiveMessage(comm);
                if (test.type == OkapiMessageType.PROCESSRESPONSE)
                {   // TODO  fix timeouts so that this thread can be killed if
                    // need be
                    resp = (ProcessResponse) test;
                    sched.processFinished(resp.finished);
                    currLoad.getAndDecrement(); // Signal a completed job
                    // TODO figure out what to do with potential Exceptions
                }
                    
            }
        }
        
        /**
         * Tells the NetHandler to stop running.
         */
        public void kill()
        {
            alive = false;
        }
    }
}
