package okapi;

import java.io.Serializable;

/**
 * This abstract class represents a process that can be distributed to any node
 * in the cluster.
 * @author Sean Kelly
 */
public abstract class OkapiProcess<E> implements Runnable, Serializable {
    private int pid; // The process ID. This is set by whatever scheduler this Process
             // is assigned to
    boolean runnable;
    
    int schedulerID; // The ID of the scheduler managing this process
    ProcState state; // The state of the process' execution
    private OkaProcessDisposal whenFinished; // Disposal technique when finished
    
    protected E retValue; // Used to return data from run
    
    /**
     * Creates a new OkapiProcess tied to a particular scheduler. The users must
     * designate whether or not the scheduler should hold completed processes
     * until retrieved by the user or dispose of them when completed.
     * @param sched The scheduler to manage the OkapiProcess.
     * @param whenFinished What to do when finished. WAIT_TIL_RECEIVED has the
     * scheduler hold the process until the user asks for it back, while DISPOSE
     * simply gets rid of the process when finished.
     */
    public OkapiProcess(Scheduler sched, OkaProcessDisposal whenFinished)
    {
        retValue = null;
        schedulerID = sched.getSchedulerCookie();
        sched.registerProcess(this);
        state = ProcState.WAITING;
        this.whenFinished = whenFinished;
    }
    
    /**
     * Creates a new OkapiProcess tied to a particular scheduler, and asks
     * the scheduler to wait for a user to retrieve processes when finished
     * by default.
     * @param sched The scheduler to manage the OkapiProcess.
     */
    public OkapiProcess(Scheduler sched)
    {
        this(sched, OkaProcessDisposal.WAIT_TIL_RECEIVED);
    }
    
    /**
     * Creates a "headless" OkapiProcess - one not assigned to a scheduler. This
     * is useful for processes that being serialized across a network.
     * @param whenFinished 
     */
    OkapiProcess(OkaProcessDisposal whenFinished)
    {
        retValue = null;
        this.whenFinished = whenFinished;
    }
    
    /**
     * Creates a "headless" OkapiProcess with the default Disposal technique.
     */
    OkapiProcess()
    {
        this(OkaProcessDisposal.WAIT_TIL_RECEIVED);
    }
    

    
    /**
     * This function is overridden such that the code within it is what is run
     * any the node where the process is scheduled. The function overriding this
     * one should place its return value into this.retValue.
     */
    @Override
    public abstract void run();
    
    /**
     * This function begins the execution of the process.
     */
    public void start()
    {
        Scheduler owner = SchedulerFactory.getInstance().getScheduler(schedulerID);
        owner.startProcess(this);
    }
    
    // Used by registering scheduler to set the pid
    void setPID(int pid)
    {
        this.pid = pid;
    }
    
    /**
     * Returns the process ID of this process, used to identify this process
     * by its scheduler.
     * @return The process ID of the process.
     */
    public int getPID()
    {
        return pid;
    }
    
    /**
     * This function returns the return value generated when execute was called.
     * @return Whatever this.retValue was set to when execute() was called. If
     * retValue was never set, this should return null.
     */
    public E getReturn()
    {
        return retValue;
    }
    
    OkaProcessDisposal disposalTechnique()
    {
        return whenFinished;
    }
    
    @Override
    public int hashCode()
    {
        return pid ^ Integer.reverse(schedulerID);
    }

    @Override
    public boolean equals(Object obj) {
        if (obj == null) {
            return false;
        }
        if (getClass() != obj.getClass()) {
            return false;
        }
        final OkapiProcess<E> other = (OkapiProcess<E>) obj;
        if (this.pid != other.pid) {
            return false;
        }
        if (this.schedulerID != other.schedulerID) {
            return false;
        }
        return true;
    }
    
    public enum ProcState
    {
        WAITING, RUNNING, ZOMBIE
    }
}
