package okapi;

import java.util.concurrent.BlockingQueue;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicInteger;

/**
 * LocalSchedulers are used to dispatch local processes. In essence, they are
 * thread pools that consume tasks given to them. As of now, this implementation
 * executes threads on a First-Come First-Served basis.
 * @author Sean Kelly
 */
public class PCThreadPool extends ProcessConsumer {
    private BlockingQueue<OkapiProcess> dispatched; // Queue of running procs
    private int maxThreads; // The upper bound on available threads
    AtomicInteger currLoad; // The number of OkapiProcesses this ThreadPool
                            // is currently processing
    protected boolean alive;
    
    Worker[] threads;
    
    /**
     * Creates a new PCThreadPool that consumes Processes from the given
     * Scheduler and is allocated the given number of threads.
     * @param sched The scheduler with which to associate this thread pool.
     * @param threadCount The number of threads allocated to the thread pool.
     */
    public PCThreadPool(Scheduler sched, int threadCount)
    {
        super(sched);
        dispatched = new LinkedBlockingQueue<OkapiProcess>();
        maxThreads = threadCount;
        currLoad = new AtomicInteger(0);
        alive = true;
        
        threads = new Worker[maxThreads];
        
        // Not sure if this should be a daemon or not.
        this.setDaemon(true);
    }
    
    @Override
    public void init()
    {
        for (int i = 0; i < maxThreads; i++)
        {
            threads[i] = new Worker(this.sched, dispatched, currLoad);
            threads[i].start();
        }
    }
    
    @Override
    public void cleanup()
    {
        alive = false; // Kill the ThreadPool
        for (int i = 0; i < maxThreads; i++) // Kill all consumer threads
            threads[i].kill();
    }

    @Override
    public boolean acceptJobs() {
        boolean ret = ((currLoad.get() - 1) <= maxThreads) && alive;
        return ret;
    }

    @Override
    public void consumeProcess(OkapiProcess proc) {
        currLoad.getAndIncrement();
        dispatched.add(proc);
    }
    
    @Override
    /**
     * Stops the ThreadPool from accepting and consuming jobs. Can be restarted,
     * as proper cleanup is not done in this case.
     */
    public void kill()
    {
        super.kill();
        this.alive = false;
    }
    

    // Workers Consume processes from the work queue given until destroyed
    private class Worker extends Thread
    {
        private Scheduler workingFor;
        private BlockingQueue<OkapiProcess> workQueue;
        private boolean alive;
        private AtomicInteger toDec;
        public Worker(Scheduler sched, BlockingQueue<OkapiProcess> work,
                AtomicInteger toDec)
        {
            super();
            workingFor = sched;
            workQueue = work;
            alive = true;
            this.toDec = toDec;
            
            
            this.setDaemon(true);
        }
        
        @Override
        public void run()
        {
            OkapiProcess toDo = null;
            while (alive)
            {
                try {
                    toDo = workQueue.poll(10, TimeUnit.MILLISECONDS);
                } catch (InterruptedException ex) {
                    // If interupted, just stop running
                    return;
                }
                if (toDo != null)
                {
                    toDo.run(); // Execute the process
                    workingFor.processFinished(toDo);
                    toDec.getAndDecrement(); // Decrement number of running jobs
                    toDo = null;
                }
            }
        }
        
        public void kill()
        {
            alive = false;
        }
    }
    
}
