package okapi;

import java.util.concurrent.TimeUnit;

/*
 * TODO give process consumers a sense of load so that scheduling can be
 * prioritized.
 */

/**
 * A process consumer belongs to a scheduler, and consumes jobs from the
 * scheduler according to internal rules.
 * @author Sean Kelly
 */
public abstract class ProcessConsumer extends Thread{
    protected Scheduler sched; // Scheduler controlling the PC
    private int pcid; // Process consumer ID
    private boolean registered; // Tells whether or not the process has been
                                // reg'd with the scheduler yet
    
    
    private boolean alive; // Tells whether or not the PC is alive
    
    /**
     * A ProcessConsumer is an abstract object that can perform the task
     * required by an OkapiProcess. The scheduler is polled for ready jobs
     * depending on 
     * @param sched 
     */
    public ProcessConsumer(Scheduler sched)
    {
        this.sched = sched;
        registered = false;
    }
    
    @Override
    public void run()
    {
        // Registration can't occur in constructor, as inheriting methods init
        // May rely on constructor data
        if (!registered)
        {
            sched.registerProcessConsumer(this);
            registered = true;
        }
        alive = true;
        OkapiProcess toDo = null;
        while (alive)
        {
            if (acceptJobs())
            {
                // Try for 10 milliseconds to get a task
                try {
                    toDo = sched.pollReadyQueue(10, TimeUnit.MILLISECONDS);
                } catch (InterruptedException ex) {}
            }
            // Consume tasks that may have arrived
            if (toDo != null)
            {
                consumeProcess(toDo);
                toDo = null;
            }
        }
    }
    
    /**
     * This function follows a child-defined set of rules to determine whether
     * it can pull more processes from the scheduler.
     * @return Whether or not more jobs should be pulled.
     */
    public abstract boolean acceptJobs();
    
    /**
     * This method is called when the ProcessConsumer has a job to do. Users
     * must <i>ensure</i> that all passed in OkapiProcesses are returned to the
     * scheduler with sched.processFinished(proc)
     * @param proc 
     */
    public abstract void consumeProcess(OkapiProcess proc);
    
    /**
     * Allows the process consumer to die. Can be restarted, as this doesn't
     * do proper cleanup.
     */
    public void kill()
    {
        alive = false;
    }
    
    void setPCID(int pcid)
    {
        this.pcid = pcid;
    }
    
    /**
     * Returns the ID that the scheduler uses to identify this PC.
     * @return The Process Consumer ID of this ProcessConsumer.
     */
    public int getPCID()
    {
        return pcid;
    }
    
    /**
     * Process consumers are given the oppurtunity to initialize members
     * after construction.
     */
    public abstract void init();
    
    public void pcleanup()
    {
        kill();
        cleanup();
    }
    
    /**
     * Process consumers should stop running threads and dispose of other
     * resources when asked to cleanup.
     */
    public abstract void cleanup();
}
