package okapi;

import java.util.Collections;
import java.util.LinkedList;
import java.util.List;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.TimeUnit;

/**
 * A Scheduler dispatches processes to slaves and manages processes that have
 * already been dispatched. Slaves can be any sort of abstract worker, such as
 * nodes in a cluster or threads in a OS process. Scheduling (as of now) is done
 * in a sort of random fashion. Processes are thrown into a FIFO structure that
 * is watched by all of the ProcessConsumers, whichever grabs the process first
 * gets it
 * @author Sean Kelly
 */
public class Scheduler {
    private CookieManager pidManager; // Manages cookies of process sent to this
                                      // Scheduler
    private CookieManager pcidManager; // Manages cookies for process consumers
    
    private int schedCookie; // A cookie used to identify processes to a certain
                             // scheduler
    // Process Lists
    private List<OkapiProcess> waitList; // Process yet to be started
    private List<OkapiProcess> procList; // List of processes being executed
    private List<OkapiProcess> zombieList; // Finished processes
    
    private BlockingQueue<OkapiProcess> readyList; // Running procs yet to be
                                                   //taken
    
    // Process Consumer List(s)
    private List<ProcessConsumer> procConsumers;
    
    Scheduler(int cookie)
    {
        pidManager = new CookieManager();
        pcidManager = new CookieManager();
        schedCookie = cookie;
        
        // Initialize Process Lists
        waitList = Collections.synchronizedList(new LinkedList<OkapiProcess>());
        procList = Collections.synchronizedList(new LinkedList<OkapiProcess>());
        zombieList = Collections.synchronizedList(
                new LinkedList<OkapiProcess>());
        
        readyList = new LinkedBlockingQueue<OkapiProcess>();
        
        // Initialize Process Consumer List(s)
        procConsumers = Collections.synchronizedList(
                new LinkedList<ProcessConsumer>());
    }
    
    /**
     * This function gets the cookie assigned to this particular scheduler. This
     * is used to identify schedulers between a potentially large base of 
     * scheduling hosts.
     * @return This servers cookie.
     */
    int getSchedulerCookie()
    {
        return schedCookie;
    }
    
    /**
     * Adds a process to the waitlist. The process becomes associated with this
     * scheduler, but it does not begin execution until the user states that it
     * should.
     * @param proc The process to be associated with this scheduler. 
     */
    void registerProcess(OkapiProcess proc)
    {
        proc.setPID(pidManager.newCookie());
        waitList.add(proc);
    }
    
    /**
     * Gives this scheduler a process that is ready to be dispatched. Processes
     * added to the scheduler will be farmed out to whatever nodes supported by
     * this scheduler.
     * @param proc 
     */
    void startProcess(OkapiProcess proc)
    {
        // Move process to ready list and dispatch
        waitList.remove(proc);
        procList.add(proc);
        proc.state = OkapiProcess.ProcState.RUNNING;
        
        // TODO Error checking if process not found
        dispatchProc(proc);
    }
    
    /**
     * Moves a process from the current process list to the zombie list.
     * Processes on the zombie list remain there until whatever service that
     * scheduled them retrieves them.
     * @param proc The process to indicate completion.
     */
    void processFinished(OkapiProcess proc)
    {
        // Assert that this process was returned to the correct Scheduler
        assert(proc.schedulerID == this.schedCookie);
        
        // Find the process on the proclist and add it to the zombielist
        procList.remove(proc);
        proc.runnable = false; // Don't allow re-run
        pidManager.releaseCookie(proc.getPID()); // Release proc cookie
        // Only place on the zombie list if users want to retrieve data
        if (proc.disposalTechnique() == OkaProcessDisposal.WAIT_TIL_RECEIVED)
            zombieList.add(proc);
        proc.state = OkapiProcess.ProcState.ZOMBIE;
    }
    
    /**
     * Dispatching actually sets the process into execution. The process sits on
     * a ready queue, which is consumed by the scheduler's process consumers.
     * @param proc The process to dispatch.
     */
    private void dispatchProc(OkapiProcess proc)
    {
        readyList.add(proc);
    }
    
    public OkapiProcess pollReadyQueue() throws InterruptedException
    {
        OkapiProcess ret = readyList.take();
        return ret;
    }
    
    public OkapiProcess pollReadyQueue(long timeout, TimeUnit unit)
                                            throws InterruptedException
    {
        OkapiProcess ret = readyList.poll(timeout, unit);
        return ret;
    }
    
    /**
     * Removes the process from the scheduler manager, freeing its cookie,
     * and cleaning up all ProcessConsumers tied to this Scheduler.
     */
    public void cleanup()
    { // TODO properly implement scheduler cleanup
        for(ProcessConsumer i : procConsumers)
            i.pcleanup();
        SchedulerFactory.getInstance().freeScheduler(schedCookie);
    }
    
    /**
     * Blocks the caller until all processes given to the scheduler are
     * completed.
     */
    public void join()
    {
        while (!(waitList.isEmpty() && procList.isEmpty()))
            Thread.yield();
    }
    
    /**
     * Blocks the caller until the given process has terminated.
     * @param proc The process to wait for.
     */
    public void join(OkapiProcess proc)
    {
        // Yield the CPU until the process becomes a zombie
        while(proc.state != OkapiProcess.ProcState.ZOMBIE)
            Thread.yield();
    }
    
    /**
     * Blocks the caller until the process finishes, then returns the completed
     * process. Returned process should not be executed again, as all of their
     * resources have been deallocated and cannot be reallocated.
     * @param proc The process to retrieve from the scheduler.
     * @return The completed process requested by the caller.
     */
    public OkapiProcess retrieveProcess(OkapiProcess proc)
    {
        // Blocks until the job is finished
        join(proc);
        zombieList.remove(proc);
        return proc;
    }
    
    /**
     * Registers a ProcessConsumer to this scheduler. When process consumers
     * registered to a particular scheduler are start()ed, they pull available
     * jobs from that scheduler.
     * @param procC 
     */
    void registerProcessConsumer(ProcessConsumer procC)
    {
        procC.setPCID(pcidManager.newCookie());
        procConsumers.add(procC);
        procC.init();
    }
}
