package okapi;

import java.io.IOException;
import java.util.HashMap;

/**
 * This class allocates and manages Schedulers that the user may need. It helps
 * to ensure that processes always report to the scheduler associated with them.
 * @author Sean Kelly
 */
public class SchedulerFactory {
    private static SchedulerFactory instance = null;
    
    public static SchedulerFactory getInstance()
    {
        if (instance == null)
            instance = new SchedulerFactory();
        return instance;
    }

    private CookieManager schedCookieManager;
    private HashMap<Integer, Scheduler> schedulers;

    private SchedulerFactory()
    {
        schedCookieManager = new CookieManager();
        schedulers = new HashMap<Integer, Scheduler>();
    }
    
    /**
     * This function is used to make a local process scheduler (Thread Pool).
     * @param numThreads The number of threads to allow the scheduler to
     * allocate.
     * @return The scheduler that has been created.
     */
    public Scheduler makeScheduler()
    {
        int cookie = schedCookieManager.newCookie();
        Scheduler sched = new Scheduler(cookie);
        schedulers.put(cookie, sched);
        return sched;
    }
    
    /**
     * This function is used to make a process scheduler that listens on the
     * given port for worker nodes that want to connect and receive jobs.
     * @param port The port on which to listen for connections
     * @return The created SchedulerServer
     * @throws IOException Throws an IOException if there is a problem opening
     * a socket on the specified port.
     */
    public SchedulerServer makeSchedulerServer(int port) throws IOException
    {
        int cookie = schedCookieManager.newCookie();
        SchedulerServer sched = new SchedulerServer(cookie, port);
        schedulers.put(cookie, sched);
        return sched;
    }
    
    /**
     * This function retrieves the scheduler attached to a particular schedID.
     * @param cookie The ID of the scheduler to retrieve.
     * @return The scheduler mapped to the given schedID.
     */
    public Scheduler getScheduler(int cookie)
    {
        return schedulers.get(cookie);
    }
    
    /**
     * Removes the given scheduler from the cache of schedulers and releases the
     * schedID back to the cookie manager.
     * @param schedID The scheduler to release.
     */
    void freeScheduler(int schedID)
    {
        schedulers.remove(schedID);
        schedCookieManager.releaseCookie(schedID);
    }
    
}
