package okapi;

import okapi.net.protocol.SchedulerPing;
import okapi.net.protocol.OkapiNetMessage;
import okapi.net.protocol.ConnectMessage;
import okapi.net.protocol.OkapiMessageType;
import okapi.net.NodeLocation;
import java.io.IOException;
import java.net.ServerSocket;
import java.net.Socket;
import java.util.List;

/**
 * A Scheduler Server listens on a particular port for connections from Slave
 * Nodes, then adds SlaveNode Process Consumers to send work to them. Scheduler
 * Servers must be assigned to a dedicated thread to work properly.
 * @author Sean Kelly
 */
public class SchedulerServer extends Scheduler implements Runnable{
    private ServerSocket lSocket; // The listening socket.

    Thread myThread; // Thread that this server runs in
    boolean alive; // Indicates whether or not the Scheduler should run.
    
    private int lport;
    
    /**
     * Constructs a new SchedulerServer that listens on the given port.
     * @param cookie The cookie assigned to this Scheduler by a Scheduler
     * Factory.
     * @param listenport The port to listen on for SlaveNode connections.
     * @throws IOException An IOException is thrown if there is a problem
     * binding to the given port.
     */
    SchedulerServer(int cookie, int listenport) throws IOException
    {
        super(cookie);
        lport = listenport;
        lSocket = new ServerSocket(listenport);
        
        alive = true;
    }
    
    public void start()
    {
        myThread = new Thread(this);
        myThread.setDaemon(true);
        myThread.start();
    }
    
    @Override
    public void run()
    {
        alive = true;
        Socket connectionHandler;
        while(alive)
        {
            try
            {   // Accept connections from SlaveNodes
                connectionHandler = lSocket.accept();
                // Handle all connections individually in seperate thread
                handleIncomingConnection(connectionHandler);
            }
            catch(IOException e) {}
        }
    }
    
    /**
     * Handles an incoming connection from a SlaveNode.
     * @param request The socket created by the server to handle the connection
     */
    private void handleIncomingConnection(Socket request)
    {
        IncomingConnectHandler handle = new IncomingConnectHandler(this, request);
        handle.start();
    }
    
    /**
     * Attaches SlaveNodes hosted at the locations specified
     * @param addrs A list of address and ports at which the SlaveNodes can be
     * found.
     */
    public void addNodes(List<NodeLocation> addrs)
    {
        OutgoingConnectHandler handler;
        for (NodeLocation i : addrs)
            handleOutgoingConnection(i);
    }
    
    /**
     * Pings a given location with a SchedulerPing pointing to the incoming
     * connection handler. Proper OkapiNodes should respond by attaching a
     * SlaveNode.
     * @param loc The location to ping.
     */
    private void handleOutgoingConnection(NodeLocation loc)
    {
        OutgoingConnectHandler handler = new OutgoingConnectHandler(loc);
        handler.start();
    }
    
    /**
     * Stops the SchedulerServer from running. Restarting the thread will bring
     * it back to life properly.
     */
    public void kill()
    {
        alive = false;
    }
    
    @Override
    public void cleanup()
    {
        super.cleanup();
        kill();
    }
    
    
    /**
     * ConnectHandlers are threads that handle connection requests made by
     * SlaveNodes. They create NetProcessConsumers that connect to the Slave
     * Nodes and send them processes to complete.
     */
    private class IncomingConnectHandler extends Thread
    {
        private SchedulerServer sched;
        private Socket request;
        /**
         * Creates a IncomingConnectHandler that starts a process consumer on the
         * given scheduler using the info
         * @param sched The SchedulerServer the client will connect to
         * @param incoming  The client's request socket
         */
        public IncomingConnectHandler(SchedulerServer sched, Socket incoming)
        {
            this.sched = sched;
            request = incoming;
        }
        
        @Override
        public void run()
        {
            OkapiNetMessage test = null;
            ConnectMessage inmsg = null;
            
            test = OkapiNetMessage.receiveMessage(request);
            if (test.type == OkapiMessageType.CONNECTMESSAGE)
                inmsg = (ConnectMessage) test;
            if (inmsg != null)
            {
                // Construct a NetProcessController for the Scheduler
                // Should automatically attach to this scheduler
                NetProcessConsumer npc = new NetProcessConsumer(sched,
                        request, inmsg);
                npc.start();
            }
        }
    }
    
    /**
     * An OutgoingConnectHandler sends a connect signal to the given SlaveNode.
     * SlaveNodes should listen for this and send back a Connect signal.
     */
    private class OutgoingConnectHandler extends Thread
    {
        private NodeLocation addr;
        public OutgoingConnectHandler(NodeLocation outAddr)
        {
            addr = outAddr;
        }
        
        @Override
        public void run()
        {
            Socket sigSocket;
            try {
                sigSocket = new Socket(addr.host, addr.port);
            } catch (IOException ex) {
                return;
            }
            SchedulerPing msg = new SchedulerPing();
            msg.port = lport;
            OkapiNetMessage.sendMessage(msg, sigSocket);
            try {
                sigSocket.close();
            } catch (IOException ex) {}
            
        }
    }
}
