package okapi;

import java.io.IOException;
import java.net.InetAddress;
import java.net.Socket;
import java.net.UnknownHostException;
import okapi.net.protocol.ConnectMessage;
import okapi.net.protocol.OkapiMessageType;
import okapi.net.protocol.OkapiNetMessage;
import okapi.net.protocol.ProcessRequest;

/**
 * A SlaveNode is a node in the cluster that binds itself to a scheduling server
 * and accepts OkapiProcesses from that server over a network connection, 
 * consumes them, and then returns them.
 * @author Sean Kelly
 */
public class SlaveNode extends Thread {
    private InetAddress servName; // Name of server to connect to
    private int port; // Port on which the server listens
    
    private Socket connection; // Connection to SchedulerServer
    private boolean alive; // Whether or not the thread should run
    
    private int availProcs; // Number of available processors
    private Scheduler procManager; // Scheduler for managing accepted jobs
    private PCThreadPool thrdPool; // Pool of threads to complete incoming jobs
    
    public SlaveNode(String serverName, int port) throws UnknownHostException
    {
        this(InetAddress.getByName(serverName), port);
    }
    
    public SlaveNode(InetAddress serverName, int port)
    {
        servName = serverName;
        this.port = port;

        availProcs = Runtime.getRuntime().availableProcessors();
        procManager = SchedulerFactory.getInstance().makeScheduler();
        thrdPool = new PCThreadPool(procManager, availProcs);
        
        alive = true;
        setDaemon(true);
    }
    
    /**
     * This function is called to establish a connection to the SchedulerServer.
     */
    private void connect() throws IOException
    {
        thrdPool.start(); // Start local thread pool
        
        // Message the schedulerserver requesting to connect to the grid
        connection = new Socket(servName, port);
        ConnectMessage msg = new ConnectMessage();
        // The only load statistics used thus far is available processors.
        msg.numJobs = availProcs + 2;
        msg.availCores = availProcs;
        OkapiNetMessage.sendMessage(msg, connection);
    }
    
    @Override
    public void run()
    {
        try { // Connect to the grid
            connect();
        } catch (IOException ex) {
            System.err.println("Error connecting to SchedulerServer.");
            return;
        }
        
        OkapiNetMessage input;
        ProcessRequest req;
        OkapiNetWrapper procWrap; // Wraps incoming processes
        while(alive)
        {   // Accept and return job requests
            input = OkapiNetMessage.receiveMessage(connection);
            if (input != null && input.type == OkapiMessageType.PROCESSREQUEST)
            {
                req = (ProcessRequest) input;
                // Wrap the process in another and send it to the scheduler
                procWrap = new OkapiNetWrapper(procManager, req.toDo, connection);
                procWrap.start();
            }
            // Cleanup if requested
            else if (input != null && input.type == OkapiMessageType.CLEANUP)
            {   // Kill the SlaveNode and cleanup its scheduler
                kill();
                procManager.join();
                procManager.cleanup();
            }
            
        }
    }

    /**
     * Signals the SlaveNode to stop working.
     */
    public void kill()
    {
        alive = false;
    }
}
